#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.DataAnalysis {
  /// <summary>
  /// Linear regression data analysis algorithm.
  /// </summary>
  [Item("TSNE", "t-distributed stochastic neighbourhood embedding projects the data in a low dimensional space to allow visual cluster identification")]
  [Creatable(CreatableAttribute.Categories.DataAnalysis, Priority = 100)]
  [StorableClass]
  public sealed class TSNEAnalysis : FixedDataAnalysisAlgorithm<IRegressionProblem> {

    #region Resultnames
    private const string ScatterPlotResultName = "Scatterplot";
    private const string DataResultName = "Projected Data";
    #endregion

    #region Parameternames
    private const string DistanceParameterName = "DistanceFunction";
    private const string PerplexityParameterName = "Perplexity";
    private const string ThetaParameterName = "Theta";
    private const string NewDimensionsParameterName = "Dimensions";
    private const string MaxIterationsParameterName = "MaxIterations";
    private const string StopLyingIterationParameterName = "StopLyingIteration";
    private const string MomentumSwitchIterationParameterName = "MomentumSwitchIteration";
    private const string InitialMomentumParameterName = "InitialMomentum";
    private const string FinalMomentumParameterName = "FinalMomentum";
    private const string EtaParameterName = "Eta";
    private const string SetSeedRandomlyParameterName = "SetSeedRandomly";
    private const string SeedParameterName = "Seed";
    private const string ClassesParameterName = "ClassNames";
    #endregion

    #region Parameterproperties
    public IFixedValueParameter<DoubleValue> PerplexityParameter
    {
      get { return Parameters[PerplexityParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public OptionalValueParameter<DoubleValue> ThetaParameter
    {
      get { return Parameters[ThetaParameterName] as OptionalValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<IntValue> NewDimensionsParameter
    {
      get { return Parameters[NewDimensionsParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IValueParameter<IDistance<RealVector>> DistanceParameter
    {
      get { return Parameters[DistanceParameterName] as IValueParameter<IDistance<RealVector>>; }
    }
    public IFixedValueParameter<IntValue> MaxIterationsParameter
    {
      get { return Parameters[MaxIterationsParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<IntValue> StopLyingIterationParameter
    {
      get { return Parameters[StopLyingIterationParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<IntValue> MomentumSwitchIterationParameter
    {
      get { return Parameters[MomentumSwitchIterationParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<DoubleValue> InitialMomentumParameter
    {
      get { return Parameters[InitialMomentumParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<DoubleValue> FinalMomentumParameter
    {
      get { return Parameters[FinalMomentumParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<DoubleValue> EtaParameter
    {
      get { return Parameters[EtaParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public IFixedValueParameter<BoolValue> SetSeedRandomlyParameter
    {
      get { return Parameters[SetSeedRandomlyParameterName] as IFixedValueParameter<BoolValue>; }
    }
    public IFixedValueParameter<IntValue> SeedParameter
    {
      get { return Parameters[SeedParameterName] as IFixedValueParameter<IntValue>; }
    }
    public IFixedValueParameter<StringValue> ClassesParameter
    {
      get { return Parameters[ClassesParameterName] as IFixedValueParameter<StringValue>; }
    }
    #endregion

    #region  Properties
    public IDistance<RealVector> Distance
    {
      get { return DistanceParameter.Value; }
    }
    public double Perplexity
    {
      get { return PerplexityParameter.Value.Value; }
    }
    public double Theta
    {
      get { return ThetaParameter.Value == null ? 0 : ThetaParameter.Value.Value; }
    }
    public int NewDimensions
    {
      get { return NewDimensionsParameter.Value.Value; }
    }
    public int MaxIterations
    {
      get { return MaxIterationsParameter.Value.Value; }
    }
    public int StopLyingIteration
    {
      get { return StopLyingIterationParameter.Value.Value; }
    }
    public int MomentumSwitchIteration
    {
      get { return MomentumSwitchIterationParameter.Value.Value; }
    }
    public double InitialMomentum
    {
      get { return InitialMomentumParameter.Value.Value; }
    }
    public double FinalMomentum
    {
      get { return FinalMomentumParameter.Value.Value; }
    }
    public double Eta
    {
      get
      {
        return EtaParameter.Value == null ? 0 : EtaParameter.Value.Value;
      }
    }
    public bool SetSeedRandomly
    {
      get { return SetSeedRandomlyParameter.Value.Value; }
    }
    public uint Seed
    {
      get { return (uint)SeedParameter.Value.Value; }
    }
    public string Classes
    {
      get { return ClassesParameter.Value.Value; }
    }

    [Storable]
    public TSNE<RealVector> tsne;
    #endregion

    #region Constructors & Cloning
    [StorableConstructor]
    private TSNEAnalysis(bool deserializing) : base(deserializing) { }
    private TSNEAnalysis(TSNEAnalysis original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) { return new TSNEAnalysis(this, cloner); }
    public TSNEAnalysis() {
      Problem = new RegressionProblem();
      Parameters.Add(new ValueParameter<IDistance<RealVector>>(DistanceParameterName, "The distance function used to differentiate similar from non-similar points", new EuclidianDistance()));
      Parameters.Add(new FixedValueParameter<DoubleValue>(PerplexityParameterName, "Perplexity-Parameter of TSNE. Comparable to k in a k-nearest neighbour algorithm. Recommended Value is Floor(number of points /3) or lower", new DoubleValue(25)));
      Parameters.Add(new OptionalValueParameter<DoubleValue>(ThetaParameterName, "Value describing how much appoximated gradients my differ from exact gradients. Set to 0 for exact calculation and in [0,1] otherwise \n CAUTION: exact calculation of forces requires building a non-sparse N*N matrix where N is the number of data points\n This may exceed memory limitations", new DoubleValue(0.1)));
      Parameters.Add(new FixedValueParameter<IntValue>(NewDimensionsParameterName, "Dimensionality of projected space (usually 2 for easy visual analysis", new IntValue(2)));
      Parameters.Add(new FixedValueParameter<IntValue>(MaxIterationsParameterName, "Maximum number of iterations for gradient descent", new IntValue(1000)));
      Parameters.Add(new FixedValueParameter<IntValue>(StopLyingIterationParameterName, "Number of iterations after which p is no longer approximated", new IntValue(250)));
      Parameters.Add(new FixedValueParameter<IntValue>(MomentumSwitchIterationParameterName, "Number of iterations after which the momentum in the gradient descent is switched", new IntValue(250)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(InitialMomentumParameterName, "The initial momentum in the gradient descent", new DoubleValue(0.5)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(FinalMomentumParameterName, "The final momentum", new DoubleValue(0.8)));
      Parameters.Add(new FixedValueParameter<DoubleValue>(EtaParameterName, "Gradient Descent learning rate", new DoubleValue(200)));
      Parameters.Add(new FixedValueParameter<BoolValue>(SetSeedRandomlyParameterName, "If the seed should be random", new BoolValue(true)));
      Parameters.Add(new FixedValueParameter<IntValue>(SeedParameterName, "The seed used if it should not be random", new IntValue(0)));
      Parameters.Add(new FixedValueParameter<StringValue>(ClassesParameterName, "name of the column specifying the class lables of each data point. \n if the lable column can not be found Training/Test is used as labels", new StringValue("none")));
    }
    #endregion

    protected override void Run() {
      var data = CalculateProjectedData(Problem.ProblemData);
      var lowDimData = new DoubleMatrix(data);
    }

    public override void Stop() {
      base.Stop();
      if (tsne != null) tsne.Running = false;
    }

    private double[,] CalculateProjectedData(IDataAnalysisProblemData problemData) {
      var DataRowNames = new Dictionary<string, List<int>>();
      var rows = new Dictionary<string, ScatterPlotDataRow>();

      if (problemData.Dataset.VariableNames.Contains(Classes)) {
        if ((problemData.Dataset as Dataset).VariableHasType<string>(Classes)) {
          var classes = problemData.Dataset.GetStringValues(Classes).ToArray();
          for (int i = 0; i < classes.Length; i++) {
            if (!DataRowNames.ContainsKey(classes[i])) DataRowNames.Add(classes[i], new List<int>());
            DataRowNames[classes[i]].Add(i); //always succeeds
          }
        } else if ((problemData.Dataset as Dataset).VariableHasType<double>(Classes)) {
          var classValues = problemData.Dataset.GetDoubleValues(Classes).ToArray();
          var max = classValues.Max() + 0.1;
          var min = classValues.Min() - 0.1;
          var contours = 8;
          for (int i = 0; i < contours; i++) {
            var name = GetContourName(i, min, max, contours);
            DataRowNames.Add(name, new List<int>());
            rows.Add(name, new ScatterPlotDataRow(name, "", new List<Point2D<double>>()));
            rows[name].VisualProperties.Color = GetHeatMapColor(i, contours);
            rows[name].VisualProperties.PointSize = i+3;
          }
          for (int i = 0; i < classValues.Length; i++) {
            DataRowNames[GetContourName(classValues[i], min, max, contours)].Add(i); //always succeeds
          }

        }


      } else {
        DataRowNames.Add("Training", problemData.TrainingIndices.ToList());
        DataRowNames.Add("Test", problemData.TestIndices.ToList());
      }

      var random = SetSeedRandomly ? new MersenneTwister() : new MersenneTwister(Seed);
      tsne = new TSNE<RealVector>(Distance, random, Results, MaxIterations, StopLyingIteration, MomentumSwitchIteration, InitialMomentum, FinalMomentum, Eta, DataRowNames, rows);
      var dataset = problemData.Dataset;
      var allowedInputVariables = problemData.AllowedInputVariables.ToArray();
      var data = new RealVector[dataset.Rows];
      for (var row = 0; row < dataset.Rows; row++) data[row] = new RealVector(allowedInputVariables.Select(col => dataset.GetDoubleValue(col, row)).ToArray());
      return tsne.Run(data, NewDimensions, Perplexity, Theta);
    }

    private static Color GetHeatMapColor(int contourNr, int noContours) {
      var q = (double)contourNr / noContours;  // q in [0,1]
      var c = q < 0.5 ? Color.FromArgb((int)(q * 2 * 255), 255, 0) : Color.FromArgb(255, (int)((1 - q) * 2 * 255), 0);
      return c;
    }
    private static string GetContourName(double value, double min, double max, int noContours) {
      var size = (max - min) / noContours;
      var contourNr = (int)((value - min) / size);
      return GetContourName(contourNr, min, max, noContours);
    }
    private static string GetContourName(int i, double min, double max, int noContours) {
      var size = (max - min) / noContours;
      return "[" + (min + i * size) + ";" + (min + (i + 1) * size) + ")";
    }

  }
}
